% notes page 10-12
% input: 
%	mu_target: the target mean, D-vector 
%	Sigma_target: the target covariance matrix, D*D
%	parameters for the Normal-Inverse-Wishart prior:
%	   mu0: D-vector
%	   kappa0: scalar
%	   nu0: scalar
%	   Lambda0: D*D
%	The effort function E() (and its derivative) is hard-coded below
% output (minimizers):
%	n: the optimal number of training points (relaxed to be in R)
%	s: sum_{i=1}^n xi
%	S: sum_{i=1}^n xi xi'  (note this is a BIG S)

function optmvn()

% the teaching goal -------------------------------------------------
D=3
mu_target = zeros(D,1)
Sigma_target = eye(D)

% the learner's initial status (prior distribution) -----------------
mu0 = zeros(D,1) 
kappa0 = 1
nu0 = 1e-5 + D-1
Lambda0 = (1e-5)*eye(D)

% finding the optimal teaching sufficient statistics ----------------
% initial values and constraints for optimization
lb=-Inf*ones(1+D+D*(D+1)/2,1);
x=zeros(1+D+D*(D+1)/2,1);
n=D+1; x(1)=n; lb(1)=0;
%s=rand(D,1); x(2:(2+D-1))=s;
%L=diag(s); % so that S=L*L' has diagonal s^2 >= s^2/n
s=n*mu_target; x(2:(2+D-1))=s;
S= n*eye(D) +n*mu_target*mu_target';
L=chol(S,'lower')
i=2+D;
for c=1:D,
  for r=c:D,
    if (c==r) lb(i)=s(c)/sqrt(n); end % S_dd=L_dd^2>=s_d^2/n
    x(i)=L(r,c); 
    i=i+1;
  end
end

f = @(x)wrapper(x,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0);
g = @(x)mycon(x,D);
% constraints: n>=0, L has nonnegative diagonal
[x, fval]=fmincon(f,x,[],[],[],[],lb,Inf*ones(size(lb)), g)
n=x(1)
s=x(2:(2+D-1))
L=zeros(D); i=2+D; for c=1:D, for r=c:D, L(r,c)=x(i); i=i+1; end; end
clear x
S=L*L'
xbar = s/n
samplecov = S/n-2/n*s*xbar'+xbar*xbar'
mu_target 
Sigma_target 

% allocating sufficient statistics into teaching examples -----------
n = round(n);
% initialize the n examples with random samples drawn from the target distribution
x0=mvnrnd(mu_target, Sigma_target, n);
% find the L2 projection of the sufficient statistics
f = @(x)allocating(x,n,D,s,S);
[x, fval]=fminunc(f, reshape(x0, n*D, 1));
x = reshape(x, n, D) % each row is a teaching example

% compute the objective after teaching with these examples ----------
s = sum(x)';
S = x'*x;
theobjective(D,n,s,S,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0)

% compare the objective after teaching with iid N(target) examples --
x0
s = sum(x0)';
S = x0'*x0;
theobjective(D,n,s,S,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0)

if (D==3)
	close all; hold on
	plot3(x(:,1),x(:,2),x(:,3),'o')
	plot3(x0(:,1),x0(:,2),x0(:,3),'r*')
	axis equal
	grid on
elseif (D==2)
	close all; hold on
	plot(x(:,1),x(:,2),'o')
	plot(x0(:,1),x0(:,2),'r*')
	axis equal
	grid on
end

end




function [f] = allocating(x, n, D, s, S)
x = reshape(x, n, D);
f = sum((sum(x)'-s).^2)+ sum(sum((x'*x-S).^2));
end

% the Cauchy-Schwarz constraints S_dd>=s_d^2/n
function [c,ceq]=mycon(x,D)
n=x(1);
s=x(2:(2+D-1));
L=zeros(D); i=2+D; for c=1:D, for r=c:D, L(r,c)=x(i); i=i+1; end; end
S=L*L';
c=s.^2/n-diag(S);
ceq=[];
end

% The teaching effort function
function [eff] = Effort(n)
c=1;
eff = c*n;
end

function [f] = wrapper(x,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0);
% unpack x into n, s, L
D=length(mu_target);
n=x(1);
s=x(2:(2+D-1));
L=zeros(D); i=2+D; for c=1:D, for r=c:D, L(r,c)=x(i); i=i+1; end; end
S=L*L';
f = theobjective(D,n,s,S,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0);
end

function [f] = theobjective(D,n,s,S,mu_target,Sigma_target,mu0,kappa0,nu0,Lambda0);
% posterior Normal-Inverse-Wishart parameters
mun = kappa0/(kappa0+n)*mu0 + 1/(kappa0+n)*s;
kappan = kappa0+n;
nun = nu0+n;
Lambdan = Lambda0 + S + kappa0*n/(kappa0+n)*mu0*mu0' - 2*kappa0/(kappa0+n)*mu0*s' - 1/(kappa0+n)*s*s';

f = nun*D/2*log(2);
for i=1:D,
  f = f + gammaln((nun+1-i)/2);
end
f = f - nun/2*log(det(Lambdan)) - D/2*log(kappan);
f = f + nun/2*log(det(Sigma_target));
f = f + 1/2*trace(inv(Sigma_target)*Lambdan);
f = f + kappan/2*(mu_target - mun)'*inv(Sigma_target)*(mu_target - mun);
f = f + Effort(n);

end
