% notes page 10-14, univariate normal with target mean and variance
% input: 
%	mu_target: the target mean (dim D=1)
%	sigma2_target: the target variance matrix
%	parameters for the Normal-Inverse-Wishart prior:
%	   mu0: D-vector
%	   kappa0: scalar
%	   nu0: scalar
%	   Lambda0: D*D
%	The effort function E() is hard-coded below
% output (minimizers):
%	n: the optimal number of training points (relaxed to be in R)
%	s: sum_{i=1}^n xi
%	S: sum_{i=1}^n xi xi'  (note this is a BIG S)

function optnormal()

% the teaching goal -------------------------------------------------
mu_target = 0;
sigma2_target = 1;

% the learner's initial status (prior distribution) -----------------
mu0 = 1;
kappa0 = 1;
nu0 = 1e-5;
Lambda0 = 1e-5;

% finding the optimal teaching sufficient statistics ----------------
% initial values and constraints for optimization
D=1;
lb=-Inf*ones(1+D+D*(D+1)/2,1);
n=2; x(1)=n; lb(1)=0;
s=rand(D,1); x(2)=s;
x(3)=sqrt(s^2/n); lb(3)=0;
f = @(x)wrapper(x,mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0);
% constraints: n>=0, L has nonnegative diagonal
[x, fval]=fmincon(f,x,[],[],[],[],lb,Inf*ones(size(lb)), @mycon)
n = x(1) 
s = x(2) 
S = x(3)^2

% allocating sufficient statistics into teaching examples -----------
n = round(n);
% initialize the n examples with random samples drawn from the target distribution
x0=normrnd(mu_target, sigma2_target, n, 1);
% find the L2 projection of the sufficient statistics
[x, fval]=fminunc(@(x)(sum(x)-s)^2+(sum(x.^2)-S)^2, x0)

% compute the objective after teaching with these examples ----------
theobjective(D,length(x),sum(x),sum(x.^2),mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0)

% compare the objective after teaching with iid N(target) examples --
x0
theobjective(D,length(x0),sum(x0),sum(x0.^2),mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0)

end


% the Cauchy-Schwarz constraint S>=s^2/n
function [c,ceq]=mycon(x)
n=x(1);
s=x(2);
S=x(3)^2;
c=s^2/n-S;
ceq=[];
end

% The teaching effort function
function [eff] = Effort(n)
c=0.5;
eff = c*n;
end

function [f] = wrapper(x,mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0);
% unpack x into n, s, L
D=length(mu_target);
n=x(1);
s=x(2:(2+D-1));
L=zeros(D); % lower triangular, S=L*L'
i=2+D;
for c=1:D,
  for r=c:D,
    L(r,c)=x(i); i=i+1;
  end
end
S=L*L';
f = theobjective(D,n,s,S,mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0);
end

function [f] = theobjective(D,n,s,S,mu_target,sigma2_target,mu0,kappa0,nu0,Lambda0);
% posterior Normal-Inverse-Wishart parameters
mun = kappa0/(kappa0+n)*mu0 + 1/(kappa0+n)*s;
kappan = kappa0+n;
nun = nu0+n;
Lambdan = Lambda0 + S + kappa0*n/(kappa0+n)*mu0*mu0' - 2*kappa0/(kappa0+n)*mu0*s' - 1/(kappa0+n)*s*s';

f = nun*D/2*log(2);
for i=1:D,
  f = f + gammaln((nun+1-i)/2);
end
f = f - nun/2*log(det(Lambdan)) - D/2*log(kappan);
f = f + nun/2*log(det(sigma2_target));
f = f + 1/2*trace(inv(sigma2_target)*Lambdan);
f = f + kappan/2*(mu_target - mun)'*inv(sigma2_target)*(mu_target - mun);
f = f + Effort(n);

end
